﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [StorableClass]
  [Item("Autoregressive TimeSeries Model", "A linear autoregressive time series model used to predict future values.")]
  public class TimeSeriesPrognosisAutoRegressiveModel : NamedItem, ITimeSeriesPrognosisModel {
    [Storable]
    public double[] Phi { get; private set; }
    [Storable]
    public double Constant { get; private set; }
    [Storable]
    public string TargetVariable { get; private set; }
    [Storable]
    public int TimeOffset { get; private set; }

    [StorableConstructor]
    protected TimeSeriesPrognosisAutoRegressiveModel(bool deserializing) : base(deserializing) { }
    protected TimeSeriesPrognosisAutoRegressiveModel(TimeSeriesPrognosisAutoRegressiveModel original, Cloner cloner)
      : base(original, cloner) {
      this.Phi = (double[])original.Phi.Clone();
      this.Constant = original.Constant;
      this.TargetVariable = original.TargetVariable;
      this.TimeOffset = original.TimeOffset;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new TimeSeriesPrognosisAutoRegressiveModel(this, cloner);
    }
    public TimeSeriesPrognosisAutoRegressiveModel(double alpha, double beta, string targetVariable)
      : base() {
      //Alpha = alpha;
      //Beta = beta;
      TargetVariable = targetVariable;
      TimeOffset = 1;
    }

    public IEnumerable<IEnumerable<double>> GetPrognosedValues(Dataset dataset, IEnumerable<int> rows, IEnumerable<int> horizons) {
      var rowsEnumerator = rows.GetEnumerator();
      var horizonsEnumerator = horizons.GetEnumerator();
      // produce a n-step forecast for all rows
      while (rowsEnumerator.MoveNext() & horizonsEnumerator.MoveNext()) {
        int row = rowsEnumerator.Current;
        int horizon = horizonsEnumerator.Current;
        double[] prognosis = new double[horizon];

        prognosis[0] = dataset.GetDoubleValue(TargetVariable, row - TimeOffset);
        for (int i = 1; i < horizon; i++) {
          //prognosis[i] = Beta * prognosis[i - 1] + Alpha;
        }
        yield return prognosis;
      }

      if (rowsEnumerator.MoveNext() || horizonsEnumerator.MoveNext())
        throw new ArgumentException("Number of elements in rows and horizon enumerations doesn't match.");
    }

    public ITimeSeriesPrognosisSolution CreateTimeSeriesPrognosisSolution(ITimeSeriesPrognosisProblemData problemData) {
      return new TimeSeriesPrognosisSolution(this, problemData);
    }

  }
}
